vim.diagnostic.config({
  -- only show virtual text for WARN and higher
  virtual_text = { severity = { min = vim.diagnostic.severity.WARN } },
  -- show diagnostic code in float if available
  float = {
    prefix = function(diagnostic, i)
      if diagnostic.code == nil then
        return string.format("[%s] ", vim.diagnostic.severity[diagnostic.severity])
      else
        return string.format("[%s] ", diagnostic.code)
      end
    end,
  },
})

local nvim_lsp = require('lspconfig')
local null_ls = require('null-ls')

local null_ls_sources = {
  null_ls.builtins.code_actions.gitsigns,
}

if vim.fn.executable("node_modules/.bin/eslint") == 1 then
  local options = {}
  -- some projects have local rules, this isn't exactly generic but is good
  -- enough
  if vim.fn.isdirectory("lint-rules") == 1 then
    options.rulePaths = { "./lint-rules" }
  end

  nvim_lsp.eslint.setup({
    settings = {
      options = options,
    },
    on_attach = function(client, bufnr)
      -- add formatting capability, the language server registers this
      -- dynamically but neovim does not support that yet
      -- https://github.com/microsoft/vscode-eslint/pull/1307
      client.server_capabilities.documentFormattingProvider = true
      client.server_capabilities.documentRangeFormattingProvider = true

      on_attach(client, bufnr)
    end,
  })
end


if vim.fn.executable("shellcheck") == 1 then
  table.insert(null_ls_sources, null_ls.builtins.diagnostics.shellcheck)
  table.insert(null_ls_sources, null_ls.builtins.code_actions.shellcheck)
end

if vim.fn.executable("deno") == 1 then
  nvim_lsp.denols.setup({
    on_attach = on_attach,
  });
else
  require('typescript').setup({
    server = {
      flags = {
        debounce_text_changes = 150,
      },
      on_attach = function(client, bufnr)
        -- mark tsserver as not having formatting available as we rely on
        -- eslint for that
        client.server_capabilities.documentFormattingProvider = false
        client.server_capabilities.documentRangeFormattingProvider = false

        on_attach(client, bufnr)

        -- override mappings for typescript
        local opts = { silent = true, buffer = bufnr }
        -- exclude import statements from reference search (may have false positives)
        vim.keymap.set('n', 'gr', '<cmd>Telescope lsp_references default_text=!import\\ <CR>', opts)
      end
    }
  })
end

if vim.fn.executable("gopls") == 1 then
  nvim_lsp.gopls.setup({
    on_attach = on_attach,
  });
elseif vim.fn.executable("gofmt") == 1 then
  table.insert(null_ls_sources, null_ls.builtins.formatting.gofmt)
end

if vim.fn.executable("solargraph") == 1 then
  nvim_lsp.solargraph.setup({
    on_attach = on_attach,
    init_options = {
      formatting = false,
    }
  })
end

local group = vim.api.nvim_create_augroup('LspFormatting', { clear = false })

null_ls.setup({
  sources = null_ls_sources,

  on_attach = function(client, bufnr)
    -- format on save
    if client.server_capabilities.documentFormattingProvider then
      for key, cmd in pairs(vim.api.nvim_get_autocmds({ group = group, buffer = bufnr })) do
        vim.api.nvim_del_autocmd(cmd.id)
      end
      vim.api.nvim_create_autocmd('BufWritePre', {
        group = group,
        buffer = bufnr,
        callback = function()
          vim.lsp.buf.format()
        end,
      })
    end
  end,
});

-- custom LSP servers
local configs = require('lspconfig.configs')

if not configs.elvish then
  configs.elvish = {
    default_config = {
      cmd = {'elvish', '--lsp'},
      filetypes = {'elvish'},
      root_dir = nvim_lsp.util.root_pattern('*.elv'),
      settings = {},
    },
  }
end

nvim_lsp.elvish.setup({
  on_attach = on_attach,
})

-- show LSP progress bar
require('fidget').setup()
